/*
 * Copyright (c) 2021 Chris Camacho (codifies -  https://bedroomcoders.co.uk/)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include "raylib.h"
#include "raymath.h"

#define RLIGHTS_IMPLEMENTATION
#include "rlights.h"

#include "nukDefs.h"
#include "nuklear.h"
#include "rayNuk.h"

// the initial window size, also used returning from full screen
#define windowWidth 1280
#define windowHeight 720

#define max( a, b ) ( ( a ) > ( b ) ? ( a ) : ( b ) )
#define min( a, b ) ( ( a ) < ( b ) ? ( a ) : ( b ) )

#define oOFF    NK_SYMBOL_CIRCLE_OUTLINE
#define oON     NK_SYMBOL_CIRCLE_SOLID

int main( void )
{
    //------------------------------------------------------------------
    // Initialization
    //------------------------------------------------------------------
    SetTraceLogLevel( LOG_ALL );
    SetConfigFlags( FLAG_WINDOW_RESIZABLE | FLAG_VSYNC_HINT | FLAG_MSAA_4X_HINT );
    InitWindow( windowWidth, windowHeight, "Raylib - template" );

    // the screen size we render to, that is then scaled
    int virtualWidth = 0, virtualHeight = 0;
    {
        int m = GetCurrentMonitor();
        virtualWidth = GetMonitorWidth( m );
        virtualHeight = GetMonitorHeight( m );
    }

    // Define the camera to look into our 3d world
    Camera camera = { 0 };
    camera.position = (Vector3){ 0.0f, 1.0f, 4.0f };
    camera.target = (Vector3){ 0.0f, 0.0f, 0.0f };
    camera.up = (Vector3){ 0.0f, 1.0f, 0.0f };
    camera.fovy = 45.0f;
    camera.type = CAMERA_PERSPECTIVE;

    // RTT target for the virtual screen
    RenderTexture2D target = LoadRenderTexture( virtualWidth, virtualHeight );
    SetTextureFilter( target.texture, FILTER_BILINEAR );
    
    // lighting shader
    Shader shader = LoadShader( "data/simpleLight.vs", "data/simpleLight.fs" );
    shader.locs[SHADER_LOC_MATRIX_MODEL] = GetShaderLocation( shader, "matModel" );
    shader.locs[SHADER_LOC_VECTOR_VIEW] = GetShaderLocation( shader, "viewPos" );

    // ambient light level
    int amb = GetShaderLocation( shader, "ambient" );
    SetShaderValue( shader, amb, (float[4]){0.2,0.2,0.2,1.0}, SHADER_UNIFORM_VEC4 );

    // make a light (max 4 but we're only using 1)
    Light light = CreateLight( LIGHT_POINT, (Vector3){ 2,4,1 }, Vector3Zero(), WHITE, shader );

    // set the shader for each material in the model
    Model model = LoadModel( "data/axis.gltf" );
    for ( int i = 0; i < model.materialCount; i++ ) {
        model.materials[i].shader = shader;
    }

    // model rotation
    Vector3 ang = { 0 };

    // setting screen size AND toggling fullscreen needs to be done separately
    int fsToggle = 0;
    ClearBackground( BLACK );
    
    //------------------------------------------------------------------
    // Gui
    //------------------------------------------------------------------
 
    struct nk_context ctx;
    struct nkdevice dev;

    // initialise the GUI
    initialiseGUI( &dev, &ctx, "data/DroidSansMono.ttf" );
    set_gui_colours( &ctx );
    
    // testing clipboard works
    char editText[41] = "HeWorlllo d!";
    int editLen = 12;

    //------------------------------------------------------------------
    // Main game loop
    //------------------------------------------------------------------
    bool exit = false;
    int rotMode = 0;
    while ( !WindowShouldClose() && !exit )    // Detect window close button or ESC key
    {
        //--------------------------------------------------------------
        // Update
        //--------------------------------------------------------------

        // handle fullscreen transition
        if ( fsToggle ) {
            fsToggle--;
            if ( !fsToggle ) {
                ToggleFullscreen();
                // need to set window size here or it upsets toggling
                if ( !IsWindowFullscreen() ) {
                    SetWindowSize( windowWidth, windowHeight );
                    SetWindowPosition( ( GetScreenWidth() - windowWidth )/2,
                        ( GetScreenHeight() - windowHeight ) / 2 );
                }
            }
        }

        // adjust mouse coordinates for scale and offset
        float screenScale = min( (float)GetScreenWidth() / virtualWidth,
                            (float)GetScreenHeight() / virtualHeight );
        
        SetMouseScale( 1.0 / screenScale, 1.0 / screenScale );
        SetMouseOffset( -( GetScreenWidth() - ( virtualWidth * screenScale ) ) * 0.5,
                        -( GetScreenHeight() - ( virtualHeight * screenScale ) ) * 0.5 );

        Vector2 mp = GetMousePosition();
        pumpInput( &ctx, mp );
        
        // do the GUI
        if ( nk_begin( &ctx, "Raylib GUI template", nk_rect(10, 10, 400, 600 ),
                     NK_WINDOW_SCALABLE | NK_WINDOW_BORDER | NK_WINDOW_MOVABLE | NK_WINDOW_CLOSABLE ) ) {


            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_STATIC, 40, 1 );
            nk_layout_row_push( &ctx, 210 );
            nk_label( &ctx, FormatText( " Screen scale %.4f", screenScale ), NK_TEXT_LEFT );
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_STATIC, 40, 2 );
            nk_layout_row_push( &ctx, 48 );
            nk_label( &ctx, "", NK_TEXT_LEFT ); // padding (fudge!?)
            nk_layout_row_push( &ctx, 210 );
            if ( nk_button_label( &ctx, "Toggle fullscreen") ) {
                fsToggle = 2;   // at least two or toggling doesn't work
                if ( !IsWindowFullscreen() ) {
                    SetWindowSize( virtualWidth, virtualHeight );
                }                
            }
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_DYNAMIC, 40, 2 );
            nk_layout_row_push( &ctx, 0.3 );
            nk_label( &ctx, " X angle:", NK_TEXT_LEFT );
            nk_layout_row_push( &ctx, 0.7 );
            nk_slider_float( &ctx, 0, &ang.x, PI*2, PI/180.f );
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_DYNAMIC, 40, 2 );
            nk_layout_row_push( &ctx, 0.3 );
            nk_label( &ctx, " Y angle:", NK_TEXT_LEFT );
            nk_layout_row_push( &ctx, 0.7 );
            nk_slider_float( &ctx, 0, &ang.y, PI*2, PI/180.f );
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_DYNAMIC, 40, 2 );
            nk_layout_row_push( &ctx, 0.3 );
            nk_label( &ctx, " Z angle:", NK_TEXT_LEFT );
            nk_layout_row_push( &ctx, 0.7 );
            nk_slider_float( &ctx, 0, &ang.z, PI*2, PI/180.f );
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_STATIC, 60, 1 );
            nk_layout_row_push( &ctx, 210 );
            nk_label( &ctx, " Rotation mode", NK_TEXT_LEFT );
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_STATIC, 0, 2 );
            nk_layout_row_push( &ctx, 48 );
            nk_label( &ctx, "", NK_TEXT_LEFT ); // padding
            nk_layout_row_push( &ctx, 210 );
            if ( nk_button_symbol_label( &ctx, ( rotMode == 0 )?oON:oOFF, "Matrix ZYX", NK_TEXT_CENTERED ) )
                    rotMode = 0;
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_STATIC, 0, 2 );
            nk_layout_row_push( &ctx, 48 );
            nk_label( &ctx, "", NK_TEXT_LEFT ); // padding
            nk_layout_row_push( &ctx, 210 );
            if ( nk_button_symbol_label( &ctx, ( rotMode == 1 )?oON:oOFF, "Matrix XYZ", NK_TEXT_CENTERED ) )
                    rotMode = 1;
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_STATIC, 0, 2 );
            nk_layout_row_push( &ctx, 48 );
            nk_label( &ctx, "", NK_TEXT_LEFT ); // padding
            nk_layout_row_push( &ctx, 210 );
            if ( nk_button_symbol_label( &ctx, ( rotMode == 2 )?oON:oOFF, "Quaternion", NK_TEXT_CENTERED ) )
                    rotMode = 2;
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_DYNAMIC, 8, 2 ); // extra padding
            //----------------------------------------------------------
            nk_layout_row_begin( &ctx, NK_DYNAMIC, 40, 2 );
            nk_layout_row_push( &ctx, 0.35 );
            nk_label( &ctx, " Clip Test", NK_TEXT_LEFT );
            nk_layout_row_push( &ctx, 0.65 );
            nk_edit_string( &ctx, NK_EDIT_ALWAYS_INSERT_MODE | NK_EDIT_FIELD | NK_EDIT_SIG_ENTER, editText, 
                                                &editLen, 40, nk_filter_ascii );            

        } else {
            exit = true;
        }
        nk_end( &ctx );

        //--------------------------------------------------------------
        // end of GUI - "normal" update stuff
        //--------------------------------------------------------------
        
        if ( rotMode==0 ) {
            model.transform = MatrixRotateZYX( ang );
        } else if ( rotMode==1 ) {
            model.transform = MatrixRotateXYZ( ang );
        } else {
            // labeled R P Y which would be Z X Y  !
            // but it's effecting XYZ axes in ZYX rotation order
            Quaternion q = QuaternionFromEuler( ang.x, ang.y, ang.z );
            model.transform = QuaternionToMatrix( q );
        }
        
        // update the camera and lighting shader
        // not actually needed unless you move the camera
        // but here so it doesn't get forgotten
        UpdateCamera( &camera );

        // position the light slightly above the camera
        light.position = camera.position;
        light.position.y += 0.1f;

        // update the light shader with the camera view position
        SetShaderValue( shader, shader.locs[SHADER_LOC_VECTOR_VIEW], &camera.position.x, SHADER_UNIFORM_VEC3 );
        UpdateLightValues( shader, light );

        //--------------------------------------------------------------
        // Draw
        //--------------------------------------------------------------
        
        BeginDrawing();

        // clears the window outside the drawing area if needed
        if ( IsWindowResized() ) {
            ClearBackground( BLACK );
        }

        // render to a texture so we can scale to the window        
        BeginTextureMode( target );
        ClearBackground( (Color){64,128,255,255} );
        
        // 3D rendering
        BeginMode3D( camera );

        DrawModel( model, (Vector3){0,0,0}, 1, WHITE );
        DrawGrid( 10, 1.0f ); 

        EndMode3D();

        // render the built GUI
        device_draw( &dev, &ctx, virtualWidth, virtualHeight, nk_vec2(1, 1), NK_ANTI_ALIASING_ON );        

        EndTextureMode();

        // draw the virtual screen on the actual one
        DrawTexturePro( target.texture, (Rectangle){ 0.0f, 0.0f,
            target.texture.width,  -target.texture.height },
            (Rectangle){ ( GetScreenWidth() - virtualWidth * screenScale ) * 0.5,
                ( GetScreenHeight() - virtualHeight * screenScale ) * 0.5,
                virtualWidth*screenScale, virtualHeight*screenScale },
                    (Vector2) { 0, 0 }, 0.0f, WHITE );


        EndDrawing();
    }

    //------------------------------------------------------------------
    // De-Initialization
    //------------------------------------------------------------------

    UnloadModel( model );
    UnloadShader( shader );

    // release the GUI
    shutDownGUI( &dev, &ctx );

    // Close window and OpenGL context
    CloseWindow();
    
    return 0;
}



