#include <stddef.h>
#include <assert.h>
#include <math.h>
#include "raylib.h"
#include "rlgl.h"

#define NK_IMPLEMENTATION
#include "nukDefs.h"
#include "nuklear.h"

#define GL_GLEXT_PROTOTYPES 1
#include <GL/glcorearb.h>

#include "rayNuk.h"



static struct nk_font_atlas atlas;
static int krpt = 0;

// I did look at using the cmd buffer to draw primatives (lines circles etc)
// but this ended up being more trouble than it was worth
// this works, however you need to directly link to GL...

#define MAX_VERTEX_MEMORY 512 * 1024
#define MAX_ELEMENT_MEMORY 128 * 1024
#define NK_SHADER_VERSION "#version 330 \n"

static void device_init( struct nkdevice *dev )
{
    // keeps compiler happy with nuklear (strict flags!)
    (void)stbrp_setup_heuristic;

    GLint status;
    static const GLchar *vertex_shader =
        NK_SHADER_VERSION
        "uniform mat4 ProjMtx;\n"
        "in vec2 Position;\n"
        "in vec2 TexCoord;\n"
        "in vec4 Color;\n"
        "out vec2 Frag_UV;\n"
        "out vec4 Frag_Color;\n"
        "void main() {\n"
        "   Frag_UV = TexCoord;\n"
        "   Frag_Color = Color;\n"
        "   gl_Position = ProjMtx * vec4(Position.xy, 0, 1);\n"
        "}\n";
    static const GLchar *fragment_shader =
        NK_SHADER_VERSION
        "precision mediump float;\n"
        "uniform sampler2D Texture;\n"
        "in vec2 Frag_UV;\n"
        "in vec4 Frag_Color;\n"
        "out vec4 Out_Color;\n"
        "void main(){\n"
        "   Out_Color = Frag_Color * texture(Texture, Frag_UV.st);\n"
        "}\n";

    nk_buffer_init_default( &dev->cmds );
    dev->prog = glCreateProgram();
    dev->vert_shdr = glCreateShader( GL_VERTEX_SHADER );
    dev->frag_shdr = glCreateShader( GL_FRAGMENT_SHADER );
    glShaderSource( dev->vert_shdr, 1, &vertex_shader, 0 );
    glShaderSource( dev->frag_shdr, 1, &fragment_shader, 0 );
    glCompileShader( dev->vert_shdr );
    glCompileShader( dev->frag_shdr );
    glGetShaderiv( dev->vert_shdr, GL_COMPILE_STATUS, &status );
    assert( status == GL_TRUE );
    glGetShaderiv( dev->frag_shdr, GL_COMPILE_STATUS, &status );
    assert( status == GL_TRUE );
    glAttachShader( dev->prog, dev->vert_shdr );
    glAttachShader( dev->prog, dev->frag_shdr );
    glLinkProgram( dev->prog );
    glGetProgramiv( dev->prog, GL_LINK_STATUS, &status );
    assert( status == GL_TRUE );

    dev->uniform_tex = glGetUniformLocation( dev->prog, "Texture" );
    dev->uniform_proj = glGetUniformLocation( dev->prog, "ProjMtx" );
    dev->attrib_pos = glGetAttribLocation( dev->prog, "Position" );
    dev->attrib_uv = glGetAttribLocation( dev->prog, "TexCoord" );
    dev->attrib_col = glGetAttribLocation( dev->prog, "Color" );

    {
        /* buffer setup */
        GLsizei vs = sizeof( struct nk_glfw_vertex );
        size_t vp = offsetof( struct nk_glfw_vertex, position );
        size_t vt = offsetof( struct nk_glfw_vertex, uv );
        size_t vc = offsetof( struct nk_glfw_vertex, col );

        glGenBuffers( 1, &dev->vbo );
        glGenBuffers( 1, &dev->ebo );
        glGenVertexArrays( 1, &dev->vao );

        glBindVertexArray( dev->vao );
        glBindBuffer( GL_ARRAY_BUFFER, dev->vbo );
        glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, dev->ebo );

        glEnableVertexAttribArray( (GLuint)dev->attrib_pos );
        glEnableVertexAttribArray( (GLuint)dev->attrib_uv );
        glEnableVertexAttribArray( (GLuint)dev->attrib_col );

        glVertexAttribPointer( (GLuint)dev->attrib_pos, 2, GL_FLOAT, GL_FALSE, vs, (void*)vp );
        glVertexAttribPointer( (GLuint)dev->attrib_uv, 2, GL_FLOAT, GL_FALSE, vs, (void*)vt );
        glVertexAttribPointer( (GLuint)dev->attrib_col, 4, GL_UNSIGNED_BYTE, GL_TRUE, vs, (void*)vc );
    }
    
    glBindTexture( GL_TEXTURE_2D, 0 );
    glBindBuffer( GL_ARRAY_BUFFER, 0 );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
    glBindVertexArray(0);
    
}

// takes coordinates for mouse as it is scaled
// to the gui surface
void pumpInput( struct nk_context *ctx, Vector2 mp )
{
    if ( IsKeyDown( KEY_LEFT ) ||
            IsKeyDown( KEY_RIGHT ) ||
            IsKeyDown( KEY_UP ) ||
            IsKeyDown( KEY_DOWN ) ) 
    {
        krpt++;
    }
    
    nk_input_begin( ctx );
    nk_input_motion( ctx, (int)mp.x, (int)mp.y );

    nk_input_button( ctx, NK_BUTTON_LEFT, (int)mp.x, (int)mp.y, IsMouseButtonDown( MOUSE_LEFT_BUTTON ) );
    nk_input_button( ctx, NK_BUTTON_MIDDLE, (int)mp.x, (int)mp.y, IsMouseButtonDown( MOUSE_MIDDLE_BUTTON ) );
    nk_input_button( ctx, NK_BUTTON_RIGHT, (int)mp.x, (int)mp.y, IsMouseButtonDown( MOUSE_RIGHT_BUTTON ) );

    nk_input_scroll( ctx, nk_vec2(0, GetMouseWheelMove() ) );

    nk_input_key( ctx, NK_KEY_DEL, IsKeyPressed( KEY_DELETE ) );
    nk_input_key( ctx, NK_KEY_ENTER, IsKeyPressed( KEY_ENTER ) );
    nk_input_key( ctx, NK_KEY_TAB, IsKeyPressed( KEY_TAB ) );
    nk_input_key( ctx, NK_KEY_BACKSPACE, IsKeyPressed( KEY_BACKSPACE ) );
    
    if ( krpt > 6 ) 
    {
        nk_input_key( ctx, NK_KEY_LEFT, IsKeyDown( KEY_LEFT ) );
        nk_input_key( ctx, NK_KEY_RIGHT, IsKeyDown( KEY_RIGHT ) );
        nk_input_key( ctx, NK_KEY_UP, IsKeyDown( KEY_UP ) );
        nk_input_key( ctx, NK_KEY_DOWN, IsKeyDown( KEY_DOWN ) );
        krpt = 0;
    }
    
    nk_input_key( ctx, NK_KEY_TEXT_LINE_START, IsKeyPressed( KEY_HOME ) );
    nk_input_key( ctx, NK_KEY_TEXT_LINE_END, IsKeyPressed( KEY_END ) );
    
    nk_input_key( ctx, NK_KEY_SHIFT, IsKeyDown( KEY_LEFT_SHIFT ) || IsKeyDown( KEY_RIGHT_SHIFT ) );
    nk_input_key( ctx, NK_KEY_CTRL, IsKeyDown( KEY_LEFT_CONTROL ) || IsKeyDown( KEY_RIGHT_CONTROL ) );
    
    if ( IsKeyDown( KEY_LEFT_CONTROL ) || IsKeyDown( KEY_RIGHT_CONTROL ) )
    {
        nk_input_key( ctx, NK_KEY_COPY, IsKeyPressed( KEY_C ) );
        nk_input_key( ctx, NK_KEY_PASTE, IsKeyPressed( KEY_V ) );
        nk_input_key( ctx, NK_KEY_CUT, IsKeyPressed( KEY_X ) );
    }
    
    int k = GetCharPressed();
    if ( k!=0 ) {
        nk_input_char( ctx, k );
    }

    nk_input_end( ctx );
}

void device_upload_atlas( struct nkdevice *dev, const void *image, int width, int height )
{
    glGenTextures( 1, &dev->font_tex );
    glBindTexture( GL_TEXTURE_2D, dev->font_tex );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, (GLsizei)width, (GLsizei)height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, image );
                 
}

void initialiseGUI( struct nkdevice *dev, struct nk_context *ctx, const char* fontName ) 
{
    device_init( dev );
    int w, h;
        // font set up
    nk_font_atlas_init_default( &atlas );
    nk_font_atlas_begin( &atlas );
    struct nk_font *font = nk_font_atlas_add_from_file( &atlas, fontName, 22, 0 );
    const void* image = nk_font_atlas_bake( &atlas, &w, &h, NK_FONT_ATLAS_RGBA32 );
    device_upload_atlas( dev, image, w, h );
    nk_font_atlas_end( &atlas, nk_handle_id( (int)dev->font_tex ), &dev->null );
    
    // initialise and set copy/paste callback
    nk_init_default( ctx, &font->handle );
    setClipCallbacks( ctx );
}

NK_INTERN void clipPaste( nk_handle usr, struct nk_text_edit *edit )
{
    (void)usr;  // unused (tell compiler we know!)
    const char *text = GetClipboardText();
    if ( text ) nk_textedit_paste( edit, text, nk_strlen( text ) );
}

NK_INTERN void clipCopy( nk_handle usr, const char *text, int len )
{
    (void)usr;
    char *str = 0;
    if ( !len ) return;
    str = (char*)malloc( (size_t)len + 1 );
    if ( !str ) return;
    memcpy( str, text, (size_t)len );
    str[len] = '\0';
    SetClipboardText( str );
    free( str );
}

void setClipCallbacks ( struct nk_context *ctx )
{
    ctx->clip.copy = clipCopy;
    ctx->clip.paste = clipPaste;
    ctx->clip.userdata = nk_handle_ptr( 0 );
}

// only here to keep it out of the way...
void set_gui_colours( struct nk_context *ctx ) 
{
    struct nk_color table[NK_COLOR_COUNT];
    table[NK_COLOR_TEXT] = nk_rgba(20, 20, 40, 255);
    table[NK_COLOR_WINDOW] = nk_rgba(202, 212, 214, 215);
    table[NK_COLOR_HEADER] = nk_rgba(137, 182, 224, 220);
    table[NK_COLOR_BORDER] = nk_rgba(140, 159, 173, 255);
    table[NK_COLOR_BUTTON] = nk_rgba(137, 182, 224, 255);
    table[NK_COLOR_BUTTON_HOVER] = nk_rgba(142/2, 187/2, 229/2, 255);
    table[NK_COLOR_BUTTON_ACTIVE] = nk_rgba(147, 192, 234, 255);
    table[NK_COLOR_TOGGLE] = nk_rgba(177, 210, 210, 255);
    table[NK_COLOR_TOGGLE_HOVER] = nk_rgba(182/2, 215/2, 215/2, 255);
    table[NK_COLOR_TOGGLE_CURSOR] = nk_rgba(137, 182, 224, 255);
    table[NK_COLOR_SELECT] = nk_rgba(177, 210, 210, 255);
    table[NK_COLOR_SELECT_ACTIVE] = nk_rgba(137, 182, 224, 255);
    table[NK_COLOR_SLIDER] = nk_rgba(177, 210, 210, 255);
    table[NK_COLOR_SLIDER_CURSOR] = nk_rgba(137, 182, 224, 245);
    table[NK_COLOR_SLIDER_CURSOR_HOVER] = nk_rgba(142, 188, 229, 255);
    table[NK_COLOR_SLIDER_CURSOR_ACTIVE] = nk_rgba(147, 193, 234, 255);
    table[NK_COLOR_PROPERTY] = nk_rgba(210, 210, 210, 255);
    table[NK_COLOR_EDIT] = nk_rgba(210, 210, 210, 225);
    table[NK_COLOR_EDIT_CURSOR] = nk_rgba(20, 20, 20, 255);
    table[NK_COLOR_COMBO] = nk_rgba(210, 210, 210, 255);
    table[NK_COLOR_CHART] = nk_rgba(210, 210, 210, 255);
    table[NK_COLOR_CHART_COLOR] = nk_rgba(137, 182, 224, 255);
    table[NK_COLOR_CHART_COLOR_HIGHLIGHT] = nk_rgba( 255, 0, 0, 255);
    table[NK_COLOR_SCROLLBAR] = nk_rgba(190, 200, 200, 255);
    table[NK_COLOR_SCROLLBAR_CURSOR] = nk_rgba(64, 84, 95, 255);
    table[NK_COLOR_SCROLLBAR_CURSOR_HOVER] = nk_rgba(70, 90, 100, 255);
    table[NK_COLOR_SCROLLBAR_CURSOR_ACTIVE] = nk_rgba(75, 95, 105, 255);
    table[NK_COLOR_TAB_HEADER] = nk_rgba(156, 193, 220, 255);
    nk_style_from_table(ctx, table);
}

void device_draw(struct nkdevice *dev, struct nk_context *ctx, int width, int height,
                 struct nk_vec2 scale, enum nk_anti_aliasing AA )
{
    GLfloat ortho[4][4] = {
        {2.0f, 0.0f, 0.0f, 0.0f},
        {0.0f, -2.0f, 0.0f, 0.0f},
        {0.0f, 0.0f, -1.0f, 0.0f},
        {-1.0f, 1.0f, 0.0f, 1.0f},
    };

    ortho[0][0] /= (GLfloat)width;
    ortho[1][1] /= (GLfloat)height;

    /* setup global state */
    glDisable( GL_CULL_FACE );
    glDisable( GL_DEPTH_TEST );
    glEnable( GL_SCISSOR_TEST );
    glActiveTexture( GL_TEXTURE0 );

    /* setup program */
    glUseProgram( dev->prog );
    glUniform1i( dev->uniform_tex, 0 );
    glUniformMatrix4fv( dev->uniform_proj, 1, GL_FALSE, &ortho[0][0] );
    {
        /* convert from command queue into draw list and draw to screen */
        const struct nk_draw_command *cmd;
        void *vertices, *elements;
        const nk_draw_index *offset = NULL;

        /* allocate vertex and element buffer */
        glBindVertexArray( dev->vao );
        glBindBuffer( GL_ARRAY_BUFFER, dev->vbo );
        glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, dev->ebo );

        glBufferData( GL_ARRAY_BUFFER, MAX_VERTEX_MEMORY, NULL, GL_STREAM_DRAW );
        glBufferData( GL_ELEMENT_ARRAY_BUFFER, MAX_ELEMENT_MEMORY, NULL, GL_STREAM_DRAW );

        /* load draw vertices & elements directly into vertex + element buffer */
        vertices = glMapBuffer( GL_ARRAY_BUFFER, GL_WRITE_ONLY );
        elements = glMapBuffer( GL_ELEMENT_ARRAY_BUFFER, GL_WRITE_ONLY );
        {
            /* fill convert configuration */
            struct nk_convert_config config;
            static const struct nk_draw_vertex_layout_element vertex_layout[] = {
                {NK_VERTEX_POSITION, NK_FORMAT_FLOAT, NK_OFFSETOF(struct nk_glfw_vertex, position)},
                {NK_VERTEX_TEXCOORD, NK_FORMAT_FLOAT, NK_OFFSETOF(struct nk_glfw_vertex, uv)},
                {NK_VERTEX_COLOR, NK_FORMAT_R8G8B8A8, NK_OFFSETOF(struct nk_glfw_vertex, col)},
                {NK_VERTEX_LAYOUT_END}
            };
            NK_MEMSET( &config, 0, sizeof(config) );
            config.vertex_layout = vertex_layout;
            config.vertex_size = sizeof( struct nk_glfw_vertex );
            config.vertex_alignment = NK_ALIGNOF( struct nk_glfw_vertex );
            config.null = dev->null;
            config.circle_segment_count = 22;
            config.curve_segment_count = 22;
            config.arc_segment_count = 22;
            config.global_alpha = 1.0f; // TODO some kind of config or changable from a function ?
            config.shape_AA = AA;
            config.line_AA = AA;

            /* setup buffers to load vertices and elements */
            {
                struct nk_buffer vbuf, ebuf;
                nk_buffer_init_fixed( &vbuf, vertices, MAX_VERTEX_MEMORY );
                nk_buffer_init_fixed( &ebuf, elements, MAX_ELEMENT_MEMORY );
                nk_convert( ctx, &dev->cmds, &vbuf, &ebuf, &config );
            }
        }
        glUnmapBuffer( GL_ARRAY_BUFFER );
        glUnmapBuffer( GL_ELEMENT_ARRAY_BUFFER );

        /* iterate over and execute each draw command */
        nk_draw_foreach( cmd, ctx, &dev->cmds ) {
            if ( !cmd->elem_count ) {
                continue;
            }
            glBindTexture( GL_TEXTURE_2D, (GLuint)cmd->texture.id );
            glScissor(
                (GLint)( cmd->clip_rect.x * scale.x ),
                (GLint)( ( height - (GLint)( cmd->clip_rect.y + cmd->clip_rect.h ) ) * scale.y ),
                (GLint)( cmd->clip_rect.w * scale.x ),
                (GLint)( cmd->clip_rect.h * scale.y ) );
            glDrawElements( GL_TRIANGLES, (GLsizei)cmd->elem_count, GL_UNSIGNED_SHORT, offset );
            offset += cmd->elem_count;
        }
        nk_clear( ctx );
    }

    /* default OpenGL state */
    glUseProgram(0);
    glBindBuffer( GL_ARRAY_BUFFER, 0 );
    glBindBuffer( GL_ELEMENT_ARRAY_BUFFER, 0 );
    glBindVertexArray(0);

    glDisable( GL_SCISSOR_TEST );
    glEnable( GL_DEPTH_TEST );
}

void device_shutdown( struct nkdevice *dev )
{
    glDetachShader( dev->prog, dev->vert_shdr );
    glDetachShader( dev->prog, dev->frag_shdr );
    glDeleteShader( dev->vert_shdr );
    glDeleteShader( dev->frag_shdr );
    glDeleteProgram( dev->prog );
    glDeleteTextures( 1, &dev->font_tex );
    glDeleteBuffers( 1, &dev->vbo );
    glDeleteBuffers( 1, &dev->ebo );
    nk_buffer_free( &dev->cmds );
}

void shutDownGUI( struct nkdevice *dev, struct nk_context *ctx ) 
{
    device_shutdown( dev );
    nk_font_atlas_clear( &atlas );
    nk_free( ctx );
}
